/*
 * GnomeCalendar widget
 * Copyright (C) 1998 the Free Software Foundation
 *
 * Author: Miguel de Icaza (miguel@kernel.org)
 */
#include <config.h>

#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <time.h>

#include <gconf/gconf.h>
#include <gconf/gconf-client.h>
#include <gnome.h>

#include "alarm.h"
#include "calendar.h"
#include "gnome-cal.h"
#include "gncal-day-panel.h"
#include "gncal-week-view.h"
#include "gnomecal-config.h"
#include "month-view.h"
#include "timeutil.h"
#include "utils-time.h"
#include "year-view.h"


static void gnome_calendar_init (GnomeCalendar *gcal);


GType
gnome_calendar_get_type (void)
{
	static GType gnome_calendar_type = 0;
	if(!gnome_calendar_type) {
		GTypeInfo gnome_calendar_info = {
			sizeof(GnomeCalendarClass),
			NULL, /* base_init */
			NULL, /* base_finalize */
			(GClassInitFunc) NULL,
			NULL, /* class_finalize */
			NULL, /* class_data */
			sizeof (GnomeCalendar),
			1,    /* n_preallocs */
			(GInstanceInitFunc) gnome_calendar_init,
		};

		gnome_calendar_type = g_type_register_static (GTK_TYPE_NOTEBOOK,
							      "GnomeCalendar",
							      &gnome_calendar_info,
							      0);
	}
	return gnome_calendar_type;
}

static void
setup_widgets (GnomeCalendar *gcal)
{
	time_t now;

	now = time (NULL);

	gcal->day_view   = gncal_day_panel_new (gcal, now);
	gcal->week_view  = gncal_week_view_new (gcal, now);
	gcal->month_view = month_view_new (gcal, now);
	gcal->year_view  = year_view_new (gcal, now);

	gtk_container_set_border_width (GTK_CONTAINER (gcal->day_view),
					GNOME_PAD_SMALL);
	gtk_container_set_border_width (GTK_CONTAINER (gcal->week_view),
					GNOME_PAD_SMALL);
	gtk_container_set_border_width (GTK_CONTAINER (gcal->month_view),
					GNOME_PAD_SMALL);
	gtk_container_set_border_width (GTK_CONTAINER (gcal->year_view),
					GNOME_PAD_SMALL);

	gcal->year_view_sw = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (gcal->year_view_sw),
					GTK_POLICY_NEVER,
					GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER (gcal->year_view_sw), gcal->year_view);
	GTK_LAYOUT (gcal->year_view)->vadjustment->step_increment = 10.0;
	gtk_adjustment_changed (GTK_ADJUSTMENT (GTK_LAYOUT (gcal->year_view)->vadjustment));

	gtk_notebook_append_page (GTK_NOTEBOOK (gcal), gcal->day_view,   gtk_label_new (_("Day View")));
	gtk_notebook_append_page (GTK_NOTEBOOK (gcal), gcal->week_view,  gtk_label_new (_("Week View")));
	gtk_notebook_append_page (GTK_NOTEBOOK (gcal), gcal->month_view, gtk_label_new (_("Month View")));
	gtk_notebook_append_page (GTK_NOTEBOOK (gcal), gcal->year_view_sw,  gtk_label_new (_("Year View")));

	gtk_widget_show_all (GTK_WIDGET (gcal));
}

static void
gnome_calendar_init (GnomeCalendar *gcal)
{
	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));

	gcal->current_day_display = gcal->current_week_display = gcal->current_month_display = 
		gcal->current_year_display = time_day_begin (time (NULL));
	gcal->cal = calendar_new (CALENDAR_INIT_ALARMS);
	setup_widgets (gcal);
}

static gchar *
gnome_calendar_get_reminder_string (iCalObject *ico)
{
	gchar *date;
	gchar *msg;
	time_t app;
	struct tm tm;

	app = ico->dalarm.trigger + ico->dalarm.offset;

	date = time_get_date_time_string (localtime_r (&app, &tm));
	msg = g_strdup_printf (_("Reminder of your appointment at %s '%s'"),
			       date, ico->summary);
	g_free (date);

	return msg;
}

static GtkWidget *
get_current_page (GnomeCalendar *gcal)
{
	gint page_num;

	page_num = gtk_notebook_get_current_page (GTK_NOTEBOOK (gcal));

	return gtk_notebook_get_nth_page (GTK_NOTEBOOK (gcal), page_num);
}

char *
gnome_calendar_get_current_view_name (GnomeCalendar *gcal)
{
	GtkWidget *page;

	g_return_val_if_fail (gcal != NULL, "dayview");
	g_return_val_if_fail (GNOME_IS_CALENDAR (gcal), "dayview");

	page = get_current_page (gcal);

	if (page == gcal->day_view)
		return "dayview";
	else if (page == gcal->week_view)
		return "weekview";
	else if (page == gcal->month_view)
		return "monthview";
	else if (page == gcal->year_view_sw)
		return "yearview";
	else
		return "dayview";
}

void
gnome_calendar_goto (GnomeCalendar *gcal, time_t new_time)
{
	GtkWidget *current;

	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));
	g_return_if_fail (new_time != -1);

	current = get_current_page (gcal);
	new_time = time_day_begin (new_time);

	if (current == gcal->day_view) {
		gncal_day_panel_set (GNCAL_DAY_PANEL (gcal->day_view), new_time);
		gcal->current_day_display = new_time;
	} else if (current == gcal->week_view) {
		gncal_week_view_set (GNCAL_WEEK_VIEW (gcal->week_view), new_time);
		gcal->current_week_display = new_time;
	} else if (current == gcal->month_view) {
		month_view_set (MONTH_VIEW (gcal->month_view), new_time);
		gcal->current_month_display = new_time;
	} else if (current == gcal->year_view_sw) {
		year_view_set (YEAR_VIEW (gcal->year_view), new_time);
		gcal->current_year_display = new_time;
	} else
		g_return_if_reached ();
}

static void
gnome_calendar_go_one_step (GnomeCalendar *gcal, gboolean step_back)
{
	GtkWidget *cp = get_current_page (gcal);
	time_t new_time;
	gint step;

	if (step_back)
		step = -1;
	else
		step = 1;

	if (cp == gcal->day_view)
		new_time = time_add_day (gcal->current_day_display, step);
	else if (cp == gcal->week_view)
		new_time = time_add_week (gcal->current_week_display, step);
	else if (cp == gcal->month_view)
		new_time = time_add_month (gcal->current_month_display, step);
	else if (cp == gcal->year_view_sw)
		new_time = time_add_year (gcal->current_year_display, step);
	else
		g_return_if_reached ();

	gnome_calendar_goto (gcal, new_time);
}

void
gnome_calendar_next (GnomeCalendar *gcal)
{
	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));

	gnome_calendar_go_one_step (gcal, FALSE);
}

void
gnome_calendar_previous (GnomeCalendar *gcal)
{
	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));

	gnome_calendar_go_one_step (gcal, TRUE);
}

void
gnome_calendar_dayjump (GnomeCalendar *gcal, time_t time)
{
	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));

	gtk_notebook_set_current_page (GTK_NOTEBOOK (gcal), 0);
	gnome_calendar_goto (gcal, time);
}

void
gnome_calendar_goto_today (GnomeCalendar *gcal)
{
	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));

	gnome_calendar_goto (gcal, time (NULL));
}

void
gnome_calendar_set_view (GnomeCalendar *gcal, const gchar *page_name)
{
	int page = 0;

	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));
	g_return_if_fail (page_name != NULL);


	if (strcmp (page_name, "dayview") == 0)
		page = 0;
	else if (strcmp (page_name, "weekview") == 0)
		page = 1;
	else if (strcmp (page_name, "monthview") == 0)
		page = 2;
	else if (strcmp (page_name, "yearview") == 0)
		page = 3;
	gtk_notebook_set_current_page (GTK_NOTEBOOK (gcal), page);
}

GtkWidget *
gnome_calendar_new (void)
{
	GnomeCalendar  *gcal;

	gcal = g_object_new (GNOME_TYPE_CALENDAR, NULL);

	return GTK_WIDGET (gcal);
}

static void
gnome_calendar_update_all (GnomeCalendar *cal, iCalObject *object, int flags)
{
	gncal_day_panel_update (GNCAL_DAY_PANEL (cal->day_view), object, flags);
	gncal_week_view_update (GNCAL_WEEK_VIEW (cal->week_view), object, flags);
	month_view_update (MONTH_VIEW (cal->month_view), object, flags);
	year_view_update (YEAR_VIEW (cal->year_view), object, flags);
}

gboolean
gnome_calendar_load (GnomeCalendar *gcal, const gchar *file)
{
	char *r;

	g_return_val_if_fail (gcal != NULL, 0);
	g_return_val_if_fail (GNOME_IS_CALENDAR (gcal), 0);
	g_return_val_if_fail (file != NULL, 0);

	if ((r = calendar_load (gcal->cal, file)) != NULL){
		printf ("Error loading calendar: %s\n", r);
		return FALSE;
	}
	gnome_calendar_update_all (gcal, NULL, 0);
	return TRUE;
}

void
gnome_calendar_add_object (GnomeCalendar *gcal, iCalObject *obj)
{
	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));
	g_return_if_fail (obj != NULL);

	calendar_add_object (gcal->cal, obj);
	gnome_calendar_update_all (gcal, obj, CHANGE_NEW);
}

void
gnome_calendar_remove_object (GnomeCalendar *gcal, iCalObject *obj)
{
	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));
	g_return_if_fail (obj != NULL);

	calendar_remove_object (gcal->cal, obj);
	gnome_calendar_update_all (gcal, obj, CHANGE_ALL);
}

void
gnome_calendar_object_changed (GnomeCalendar *gcal, iCalObject *obj, int flags)
{
	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));
	g_return_if_fail (obj != NULL);

	gcal->cal->modified = TRUE;

	gnome_calendar_update_all (gcal, obj, flags);
	calendar_object_changed (gcal->cal, obj, flags);
}

static int
max_open_files (void)
{
        static int files;

        if (files)
                return files;

        files = sysconf (_SC_OPEN_MAX);
        if (files != -1)
                return files;
#ifdef OPEN_MAX
        return files = OPEN_MAX;
#else
        return files = 256;
#endif
}

static void
execute (char *command, int close_standard)
{
	struct sigaction ignore, save_intr, save_quit;
	int status = 0, i;
	pid_t pid;

	ignore.sa_handler = SIG_IGN;
	sigemptyset (&ignore.sa_mask);
	ignore.sa_flags = 0;

	sigaction (SIGINT, &ignore, &save_intr);
	sigaction (SIGQUIT, &ignore, &save_quit);

	if ((pid = fork ()) < 0){
		fprintf (stderr, "\n\nfork () = -1\n");
		return;
	}
	if (pid == 0){
		pid = fork ();
		if (pid == 0){
			const int top = max_open_files ();
			sigaction (SIGINT,  &save_intr, NULL);
			sigaction (SIGQUIT, &save_quit, NULL);

			for (i = (close_standard ? 0 : 3); i < top; i++)
				close (i);

			/* FIXME: As an excercise to the reader, copy the
			 * code from mc to setup shell properly instead of
			 * /bin/sh.  Yes, this comment is larger than a cut and paste.
			 */
			execl ("/bin/sh", "/bin/sh", "-c", command, (char *) 0);

			_exit (127);
		} else {
			_exit (127);
		}
	}
	wait (&status);
	sigaction (SIGINT,  &save_intr, NULL);
	sigaction (SIGQUIT, &save_quit, NULL);
}

void
mail_notify (char *mail_address, char *text, time_t app_time)
{
	pid_t pid;
	int   p [2];
	char *command;

	pipe (p);
	pid = fork ();
	if (pid == 0){
		int dev_null;

		dev_null = open ("/dev/null", O_RDWR);
		dup2 (p [0], 0);
		dup2 (dev_null, 1);
		dup2 (dev_null, 2);
		close (p [1]);
		execl ("/usr/lib/sendmail", "/usr/lib/sendmail",
		       mail_address, NULL);
		_exit (127);
	}
	command = g_strconcat ("To: ", mail_address, "\n",
				  "Subject: ", _("Reminder of your appointment at "),
				  ctime (&app_time), "\n\n", text, "\n", NULL);
	close (p [0]);
	write (p [1], command, strlen (command));
 	close (p [1]);
	g_free (command);
}

static void
stop_beeping (GObject* object, gpointer data)
{
	guint timer_tag, beep_tag;
	timer_tag = GPOINTER_TO_INT (g_object_get_data (object, "timer_tag"));
	beep_tag  = GPOINTER_TO_INT (g_object_get_data (object, "beep_tag"));
	
	if (beep_tag > 0) {
		gtk_timeout_remove (beep_tag);
		g_object_set_data (object, "beep_tag", GINT_TO_POINTER (0));
	}
	if (timer_tag > 0) {
		gtk_timeout_remove (timer_tag);
		g_object_set_data (object, "timer_tag", GINT_TO_POINTER (0));
	}
}

static gint
start_beeping (gpointer data)
{
	gdk_beep ();

	return TRUE;
}

static gint
timeout_beep (gpointer data)
{
	stop_beeping (data, NULL);
	return FALSE;
}

static void
alarm_dialog_response (GtkDialog *dialog, gint button, gpointer data)
{
	time_t now, diff;
	CalendarAlarm *which = g_object_get_data (G_OBJECT (dialog),
						  "cal_alarm");
	iCalObject *ico = (iCalObject*) data;
	
	gboolean is_audio = (&ico->aalarm == which);

	switch (button) {
	case 1: {
		GConfClient *gconf;
		gint snooze_secs;

		if (is_audio)
			stop_beeping (G_OBJECT (dialog), NULL);

		gconf = gconf_client_get_default ();
		snooze_secs = gconf_client_get_int (gconf, CONFIG_ALARM_SNOOZE,
						    NULL);
		snooze_secs = CLAMP (snooze_secs,
				     DEFAULT_ALARM_SNOOZE_MIN,
				     DEFAULT_ALARM_SNOOZE_MAX);
		g_object_unref (G_OBJECT (gconf));

		now = time (NULL);
		diff = now - which->trigger;
		which->trigger = which->trigger + diff + snooze_secs;
		which->offset  = which->offset  - diff - snooze_secs;
		alarm_add (which, &calendar_notify, data);
		break;
	}
	default:
		if (is_audio)
			stop_beeping (G_OBJECT (dialog), NULL);
		break;
	}
}

static gboolean
alarm_dialog_closed (GtkDialog *dialog, gpointer data)
{
	CalendarAlarm *which = g_object_get_data (G_OBJECT (dialog),
						  "cal_alarm");
	iCalObject *ico = (iCalObject*) data;
	
	gboolean is_audio = (&ico->aalarm == which);

	if (is_audio)
		stop_beeping (G_OBJECT (dialog), NULL);
	
	return FALSE;
}

void
calendar_notify (time_t activation_time, CalendarAlarm *which, void *data)
{
	iCalObject *ico = data;
	guint beep_tag, timer_tag;

	if (&ico->aalarm == which){

		GConfClient *gconf;
		GtkWidget *dialog;
		gchar *msg;
		gint timeout;

		gconf = gconf_client_get_default ();
		timeout = gconf_client_get_int (gconf, CONFIG_ALARM_TIMEOUT, NULL);
		if (timeout)
			timeout = CLAMP (timeout,
					 DEFAULT_ALARM_TIMEOUT_MIN,
					 DEFAULT_ALARM_TIMEOUT_MAX);
		g_object_unref (G_OBJECT (gconf));

		msg = gnome_calendar_get_reminder_string (ico);

		dialog = gtk_message_dialog_new (NULL, 0,
						 GTK_MESSAGE_INFO,
						 GTK_BUTTONS_NONE,
						 "%s", msg);

		g_free (msg);

		gtk_dialog_add_buttons (GTK_DIALOG (dialog),
					_("Snooze"), 1,
					GTK_STOCK_OK, 0,
					NULL);

		beep_tag = gtk_timeout_add (1000, start_beeping, NULL);
		if (timeout >= 1 && timeout <= 3600)
			timer_tag = gtk_timeout_add (timeout * 1000, 
						     timeout_beep, dialog);
		else
			timer_tag = 0;

		g_object_set_data (G_OBJECT (dialog), "timer_tag",
				   GINT_TO_POINTER (timer_tag));
		g_object_set_data (G_OBJECT (dialog), "beep_tag",
				   GINT_TO_POINTER (beep_tag));
		g_object_set_data (G_OBJECT (dialog), "cal_alarm", which);

		g_signal_connect (G_OBJECT (dialog), "response",
				  G_CALLBACK (alarm_dialog_response), data);
		g_signal_connect (G_OBJECT (dialog), "close",
				  G_CALLBACK (alarm_dialog_closed), data);

		gtk_widget_show (dialog);

	} else if (&ico->palarm == which){

		execute (ico->palarm.data, 0);

	} else if (&ico->malarm == which){

		time_t app = ico->malarm.trigger + ico->malarm.offset;

		mail_notify (ico->malarm.data, ico->summary, app);

	} else if (&ico->dalarm == which) {

		GConfClient *gconf;
		GtkWidget *dialog;
		gchar *msg;

		gconf = gconf_client_get_default ();

		if (gconf_client_get_bool (gconf, CONFIG_ALARM_BEEP, NULL))
			gdk_beep ();

		g_object_unref (G_OBJECT (gconf));

		msg = gnome_calendar_get_reminder_string (ico);

		dialog = gtk_message_dialog_new (NULL, 0,
						 GTK_MESSAGE_INFO,
						 GTK_BUTTONS_NONE,
						 msg);

		g_free (msg);

		gtk_dialog_add_buttons (GTK_DIALOG (dialog),
					_("Snooze"), 1,
					GTK_STOCK_OK, 0,
					NULL);

		g_signal_connect (G_OBJECT (dialog), "response",
				  G_CALLBACK (alarm_dialog_response), data);

		g_object_set_data (G_OBJECT (dialog), "cal_alarm", which);

		gtk_widget_show (dialog);
	}
}

/*
 * called from the calendar_iterate routine to mark the days of a GtkCalendar
 */
static int
mark_gtk_calendar_day (iCalObject *obj, time_t start, time_t end, void *c)
{
	GtkCalendar *gtk_cal = c;
	struct tm tm_s;
	time_t t, day_end;

	tm_s = *localtime (&start);
	day_end = time_day_end (end);

	for (t = start; t < day_end; t += 60*60*24){
		time_t new = mktime (&tm_s);
		struct tm tm_day;

		tm_day = *localtime (&new);
		gtk_calendar_mark_day (gtk_cal, tm_day.tm_mday);
		tm_s.tm_mday++;
	}
	return TRUE;
}

/*
 * Tags the dates with appointments in a GtkCalendar based on the
 * GnomeCalendar contents
 */
void
gnome_calendar_tag_calendar (GnomeCalendar *cal, GtkCalendar *gtk_cal)
{
	time_t month_begin, month_end;
	struct tm tm;

	g_return_if_fail (cal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (cal));
	g_return_if_fail (gtk_cal != NULL);
	g_return_if_fail (GTK_IS_CALENDAR (gtk_cal));

	/* compute month_begin */
	tm.tm_hour = 0;
	tm.tm_min  = 0;
	tm.tm_sec  = 0;
	tm.tm_mday = 1; /* setting this to zero is a no-no; it will set mktime back to the end of the
					   previous month, which may be 28,29,30; this may chop some days from the calendar */
	tm.tm_mon  = gtk_cal->month;
	tm.tm_year = gtk_cal->year - 1900;
	tm.tm_isdst= -1;

	month_begin = mktime (&tm);
	tm.tm_mon++;
	month_end   = mktime (&tm);

	gtk_calendar_freeze (gtk_cal);
	gtk_calendar_clear_marks (gtk_cal);
	calendar_iterate (cal->cal, month_begin, month_end, mark_gtk_calendar_day, gtk_cal);
	gtk_calendar_thaw (gtk_cal);
}

time_t
gnome_calendar_current_display (GnomeCalendar *gcal)
{
	GtkWidget *page;

	g_return_val_if_fail (gcal != NULL, -1);
	g_return_val_if_fail (GNOME_IS_CALENDAR (gcal), -1);

	page = get_current_page (gcal);
	if (page == gcal->day_view)
		return gcal->current_day_display;
	if (page == gcal->week_view)
		return gcal->current_week_display;
	if (page == gcal->month_view)
		return gcal->current_month_display;
	if (page == gcal->year_view_sw)
		return gcal->current_year_display;

	g_assert_not_reached();
	return -1;
}
	
