/* Functions to mark calendars
 *
 * Copyright (C) 1998 Red Hat Software, Inc.
 *
 * Author: Federico Mena <federico@nuclecu.unam.mx>
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gconf/gconf.h>
#include <gconf/gconf-client.h>

#include "gnome-cal.h"
#include "gnomecal-config.h"
#include "mark.h"
#include "timeutil.h"


/*
 * Forward Declarations
 */

/* Marks a day specified by index, not by day number */
static void mark_month_item_index (GnomeMonthItem *mitem, int index, CalendarObject *obj);


/* Frees the specified data when an object is destroyed */
static void
free_data (GtkObject *object, gpointer data)
{
	g_free (data);
}

/* If the array of "marked" attributes for the days in a a month item has not been created yet, this
 * function creates the array and clears it.  Otherwise, it just returns the existing array.
 */
static gboolean *
get_attributes (GnomeMonthItem *mitem)
{
	gboolean *attrs;

	attrs = g_object_get_data (G_OBJECT (mitem), "day_mark_attributes");

	if (!attrs) {
		attrs = g_new0 (gboolean, 42);
		g_object_set_data (G_OBJECT (mitem),
				   "day_mark_attributes", attrs);
		g_signal_connect (G_OBJECT (mitem), "destroy",
				  G_CALLBACK (free_data), attrs);
	}

	return attrs;
}

/* In the month item, marks all the days that are touched by the specified time span.  Assumes that
 * the time span is completely contained within the month.  The array of day attributes is modified
 * accordingly.
 */
static void
mark_event_in_month (GnomeMonthItem *mitem, time_t start, time_t end, CalendarObject *obj)
{
	struct tm tm;
	int day_index;
	
	tm = *localtime (&start);

	for (; start <= end; start += 60 * 60 * 24) {
		mktime (&tm); /* normalize the time */

		/* Figure out the day index that corresponds to this time */

		day_index = gnome_month_item_day2index (mitem, tm.tm_mday);
		g_assert (day_index >= 0);

		/* Mark the day box */

		mark_month_item_index (mitem, day_index, obj);

		/* Next day */

		tm.tm_mday++;
	}
}

void
mark_month_item (GnomeMonthItem *mitem, Calendar *cal)
{
	time_t month_begin, month_end;
	GList *list, *l;
	CalendarObject *co;
	guint year, month;

	g_return_if_fail (mitem != NULL);
	g_return_if_fail (GNOME_IS_MONTH_ITEM (mitem));
	g_return_if_fail (cal != NULL);

	g_object_get (G_OBJECT (mitem),
		      "year",  &year,
		      "month", &month,
		      NULL);

	month_begin = time_month_begin (time_from_day (year, month, 1));
	month_end   = time_month_end (month_begin);

	list = calendar_get_events_in_range (cal, month_begin, month_end);

	for (l = list; l; l = l->next) {
		co = l->data;

		/* We clip the event's start and end times to the month's limits */

		mark_event_in_month (mitem, MAX (co->ev_start, month_begin), MIN (co->ev_end, month_end), co);
	}

	calendar_destroy_event_list (list);
}

static gchar *
mark_get_color (GConfClient *gconf, const gchar *key, const gchar *def)
{
	gchar *color;

	color = gconf_client_get_string (gconf, key, NULL);
	if (!color || !pango_color_parse (NULL, color)) {
		g_free (color);
		color = g_strdup (def);
	}

	return color;
}


static void
mark_month_item_index (GnomeMonthItem *mitem, int index, CalendarObject *obj)
{
	GConfClient *gconf;
	gchar *color;
	gboolean *attrs;

	g_return_if_fail (mitem != NULL);
	g_return_if_fail (GNOME_IS_MONTH_ITEM (mitem));
	g_return_if_fail ((index >= 0) && (index < 42));

	attrs = get_attributes (mitem);

	attrs[index] = TRUE;

	gconf = gconf_client_get_default ();

	if (obj && obj->ico->allday) {
		GnomeCanvasItem *item_label;
		item_label = gnome_month_item_num2child (mitem, GNOME_MONTH_ITEM_DAY_LABEL + index);
		color = mark_get_color (gconf, CONFIG_COLOR_EVENT_DAYS, DEFAULT_COLOR_EVENT_DAYS);
		gnome_canvas_item_set (item_label, "fill_color", color, NULL);
	} else {
		GnomeCanvasItem *item;
		item = gnome_month_item_num2child (mitem, GNOME_MONTH_ITEM_DAY_BOX + index);
		color = mark_get_color (gconf, CONFIG_COLOR_APPOINTMENTS, DEFAULT_COLOR_APPOINTMENTS);
		gnome_canvas_item_set (item, "fill_color", color, NULL);
		g_object_set_data (G_OBJECT (item), "bg_type", GINT_TO_POINTER (1));
	}

	g_free (color);
	g_object_unref (G_OBJECT (gconf));
}

void
unmark_month_item (GnomeMonthItem *mitem)
{
	int i;
	gboolean *attrs;
	GnomeCanvasItem *item;
	GConfClient *gconf;
	gchar *color;

	g_return_if_fail (mitem != NULL);
	g_return_if_fail (GNOME_IS_MONTH_ITEM (mitem));

	attrs = get_attributes (mitem);

	/* Find marked days and unmark them by turning off their marked
	 * attribute flag and changing the color.
	 */

	gconf = gconf_client_get_default ();
	color = mark_get_color (gconf, CONFIG_COLOR_EMPTY_DAYS, DEFAULT_COLOR_EMPTY_DAYS);
	g_object_unref (G_OBJECT (gconf));

	for (i = 0; i < 42; i++) {
		if (attrs[i]) {
			attrs[i] = FALSE;

			item = gnome_month_item_num2child (mitem, GNOME_MONTH_ITEM_DAY_BOX + i);
			gnome_canvas_item_set (item, "fill_color", color, NULL);
		}
	}

	g_free (color);
}
