/* Card List Widget Object
 * Copyright (C) 2002  Sebastian Rittau <srittau@jroger.in-berlin.de>
 *
 * $Id: cardlist-widget.c,v 1.1.2.5 2002/07/21 13:38:24 srittau Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gconf/gconf.h>
#include <gconf/gconf-client.h>

#include <gnome.h>

#include <mimedir/mimedir-vcard.h>

#include "cardlist-headers.h"
#include "cardlist-widget.h"
#include "gnomecard-config.h"


/* FIXME: list sortable, drag & drop */


struct _CardListWidgetPriv {
	guint notify;
};


static GtkTreeViewClass *parent_class = NULL;


static void cardlist_widget_class_init		(CardListWidgetClass *klass);
static void cardlist_widget_init		(CardListWidget *widget);
static void cardlist_widget_dispose		(GObject *object);

static void cardlist_widget_setup		(CardListWidget *widget);

static void cardlist_widget_card_changed	(MIMEDirVCard *card, gpointer data);

/*
 * Class and Object Management
 */

GType
cardlist_widget_get_type (void)
{
	static GType cardlist_widget_type = 0;

	if (!cardlist_widget_type) {
		static const GTypeInfo cardlist_widget_info = {
			sizeof (CardListWidgetClass),
			NULL, /* base_init */
			NULL, /* base_finalize */
			(GClassInitFunc) cardlist_widget_class_init,
			NULL, /* class_finalize */
			NULL, /* class_data */
			sizeof (CardListWidget),
			1,    /* n_preallocs */
			(GInstanceInitFunc) cardlist_widget_init,
		};

		cardlist_widget_type = g_type_register_static (GTK_TYPE_TREE_VIEW,
							       "CardListWidget",
							       &cardlist_widget_info,
							       0);
	}

	return cardlist_widget_type;
}

static void
cardlist_widget_class_init (CardListWidgetClass *klass)
{
	GObjectClass *gobject_class;

	g_return_if_fail (klass != NULL);
	g_return_if_fail (CARDLIST_IS_WIDGET_CLASS (klass));

	gobject_class = G_OBJECT_CLASS (klass);

	gobject_class->dispose = cardlist_widget_dispose;

	parent_class = g_type_class_peek_parent (klass);
}

static void
cardlist_widget_init (CardListWidget *widget)
{
	g_return_if_fail (widget != NULL);
	g_return_if_fail (CARDLIST_IS_WIDGET (widget));

	widget->priv = g_new0 (CardListWidgetPriv, 1);

	cardlist_widget_setup (widget);
}

static void
cardlist_widget_dispose (GObject *object)
{
	CardListWidget *widget;

	g_return_if_fail (object != NULL);
	g_return_if_fail (CARDLIST_IS_WIDGET (object));

	widget = CARDLIST_WIDGET (object);

	if (widget->priv) {
		/* FIXME: disconnect */

		if (widget->priv->notify) {
			GConfClient *gconf;

			gconf = gconf_client_get_default ();

			gconf_client_notify_remove (gconf, widget->priv->notify);
			widget->priv->notify = 0;

			g_object_unref (G_OBJECT (gconf));
		}

		g_free (widget->priv);
		widget->priv = NULL;
	}

	G_OBJECT_CLASS (parent_class)->dispose (object);
}

/*
 * Internal Methods
 */

static void
cardlist_widget_hide_all_columns (CardListWidget *widget)
{
	CardListHeader *header;
	gint i;

	for (i = 0, header = cardlist_header_list; header->name != NULL; i++, header++) {
		GtkTreeViewColumn *column;
		column = gtk_tree_view_get_column (GTK_TREE_VIEW (widget), i);
		gtk_tree_view_column_set_visible (column, FALSE);
	}
}

static void
cardlist_widget_show_column_by_name (CardListWidget *widget, const gchar *col)
{
	GtkTreeViewColumn *tree_column;
	CardListHeader *header;
	gint i, column = -1;

	for (i = 0, header = cardlist_header_list; header->name != NULL; i++, header++) {
		if (!g_ascii_strcasecmp (col, header->name)) {
			column = i;
			break;
		}
	}

	if (column < 0) {
		g_warning (_("'%s' is not a valid column name"), col);
		return;
	}

	tree_column = gtk_tree_view_get_column (GTK_TREE_VIEW (widget), column);
	gtk_tree_view_column_set_visible (tree_column, TRUE);
}

static void
cardlist_widget_show_columns (CardListWidget *widget, GSList *list)
{
	GSList *item;

	/* Hide all columns */

	cardlist_widget_hide_all_columns (widget);

	/* Show selected columns */

	if (list) {
		for (item = list; item != NULL; item = g_slist_next (item))
			cardlist_widget_show_column_by_name (widget, (const gchar *) item->data);
	} else
		cardlist_widget_show_column_by_name (widget, "FullName");
}


static GtkTreeIter
cardlist_widget_internal_append (CardListWidget *widget, MIMEDirVCard *vcard)
{
	GtkTreeModel *model;
	GtkTreeIter   iter;
	gulong signal;

	model = gtk_tree_view_get_model (GTK_TREE_VIEW (widget));

	signal = g_signal_connect (G_OBJECT (vcard), "changed", G_CALLBACK (cardlist_widget_card_changed), widget);

	gtk_list_store_append (GTK_LIST_STORE (model), &iter);
	gtk_list_store_set (GTK_LIST_STORE (model), &iter, 0, vcard, 1, signal, -1);

	return iter;
}


static void
cardlist_widget_internal_remove (CardListWidget *widget, MIMEDirVCard *vcard)
{
	GtkTreeModel *model;
	GtkTreeIter iter;

	model = gtk_tree_view_get_model (GTK_TREE_VIEW (widget));

	if (gtk_tree_model_get_iter_root (model, &iter)) {
		do {
			MIMEDirVCard *vc;
			gulong signal;

			gtk_tree_model_get (model, &iter, 0, &vc, 1, &signal, -1);
			g_assert (vc != NULL && MIMEDIR_IS_VCARD (vc));

			if (vcard == vc) {
				g_signal_handler_disconnect (G_OBJECT (widget), signal);

				gtk_list_store_remove (GTK_LIST_STORE (model), &iter);

				return;
			}

		} while (gtk_tree_model_iter_next (model, &iter));
	}
}

/*
 * Callbacks
 */

static void
cardlist_widget_render (GtkTreeViewColumn *tree_column,
			GtkCellRenderer   *renderer,
			GtkTreeModel      *model,
			GtkTreeIter       *iter,
			gpointer           data)
{
	MIMEDirVCard *card;
	const gchar *property;
	gchar *string;

	g_return_if_fail (data != NULL);

	property = (const gchar *) data;

	gtk_tree_model_get (model, iter, 0, &card, -1);

	g_object_get (G_OBJECT (card), property, &string, NULL);
	g_object_set (G_OBJECT (renderer), "text", string ? string : "", NULL);

	g_free (string);
}

static void
cardlist_widget_config_changed (GConfClient *gconf,
				guint        cnxn_id,
				GConfEntry  *entry,
				gpointer     data)
{
	CardListWidget *widget;
        const gchar *key;

        g_return_if_fail (entry != NULL);
        g_return_if_fail (data != NULL);
        g_return_if_fail (CARDLIST_IS_WIDGET (data));

	widget = CARDLIST_WIDGET (data);

        key = gconf_entry_get_key (entry);

	if (!strcmp (key, CONFIG_UI_COLUMN_HEADERS)) {
		GConfValue *value = gconf_entry_get_value (entry);

		if (value->type == GCONF_VALUE_LIST && gconf_value_get_list_type (value) == GCONF_VALUE_STRING) {
			GSList *list, *new_list = NULL;

			/* Show selected columns */

			list = gconf_value_get_list (value);
			for (; list != NULL; list = g_slist_next (list)) {
				const gchar *s;
				s = gconf_value_get_string ((GConfValue *) list->data);
				new_list = g_slist_append (new_list, (gpointer) s);
			}
			cardlist_widget_show_columns (widget, new_list);
			g_slist_free (new_list);
		}
	}
}


static gboolean
cardlist_widget_card_changed_fe (GtkTreeModel *model,
				 GtkTreePath  *path,
				 GtkTreeIter  *iter,
				 gpointer      data)
{
	MIMEDirVCard *card, *curitem;

	g_return_val_if_fail (data != NULL, TRUE);
	g_return_val_if_fail (MIMEDIR_IS_VCARD (data), TRUE);

	card = MIMEDIR_VCARD (data);

        gtk_tree_model_get (model, iter, 0, &curitem, -1);

        g_assert (curitem != NULL);
        g_assert (MIMEDIR_IS_VCARD (curitem));

        if (card == curitem) {
                gtk_tree_model_row_changed (model, path, iter);

                return TRUE;
        }

        return FALSE;
}


static void
cardlist_widget_card_changed (MIMEDirVCard *card, gpointer data)
{
	CardListWidget *widget;
	GtkTreeModel *model;

	g_return_if_fail (card != NULL);
	g_return_if_fail (MIMEDIR_IS_VCARD (card));
	g_return_if_fail (data != NULL);
	g_return_if_fail (CARDLIST_IS_WIDGET (data));

	widget = CARDLIST_WIDGET (data);
        model = gtk_tree_view_get_model (GTK_TREE_VIEW (widget));

	gtk_tree_model_foreach (model, cardlist_widget_card_changed_fe, card);
}

/*
 * Setup
 */

static void
cardlist_widget_setup_gconf (CardListWidget *widget)
{
	CardListWidgetPriv *priv = widget->priv;
	GConfClient *gconf;
	GSList *list;

	cardlist_widget_hide_all_columns (widget);

	gconf = gconf_client_get_default ();

	gconf_client_add_dir (gconf, CONFIG_UI_DIR, GCONF_CLIENT_PRELOAD_ONELEVEL, NULL);

	list = gconf_client_get_list (gconf, CONFIG_UI_COLUMN_HEADERS,
				      GCONF_VALUE_STRING, NULL);

	/* FIXME: free the result? How? */

	cardlist_widget_show_columns (widget, list);

	/* Notification */

	priv->notify = gconf_client_notify_add (gconf, CONFIG_UI_COLUMN_HEADERS,
						cardlist_widget_config_changed, widget,
						NULL, NULL);

	g_object_unref (G_OBJECT (gconf));
}

static void
cardlist_widget_setup (CardListWidget *widget)
{
	CardListWidgetPriv *priv;
	GtkListStore *store;
	CardListHeader *header;
	gint i;

	priv = widget->priv;

	store = gtk_list_store_new (2, G_TYPE_OBJECT, G_TYPE_ULONG);

	gtk_tree_view_set_model (GTK_TREE_VIEW (widget), GTK_TREE_MODEL (store));

	for (i = 0, header = cardlist_header_list; header->name != NULL; i++, header++) {
		GtkCellRenderer *renderer;

		renderer = gtk_cell_renderer_text_new ();
		gtk_tree_view_insert_column_with_data_func (GTK_TREE_VIEW (widget),
							    i,
							    _(header->title),
							    renderer,
							    cardlist_widget_render,
							    header->property, NULL);
	}

	cardlist_widget_setup_gconf (widget);
}

/*
 * Public Methods
 */

GtkWidget *
cardlist_widget_new (void)
{
	CardListWidget *widget;

	widget = g_object_new (CARDLIST_TYPE_WIDGET, NULL);

	return GTK_WIDGET (widget);
}


MIMEDirVCard *
cardlist_widget_get_current_card (CardListWidget *widget)
{
	GtkTreeSelection *selection;
	GtkTreeModel *model;
	GtkTreeIter iter;
	MIMEDirVCard *vcard;

	g_return_val_if_fail (widget != NULL, NULL);
	g_return_val_if_fail (CARDLIST_IS_WIDGET (widget), NULL);

	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (widget));
	if (!gtk_tree_selection_get_selected (selection, &model, &iter))
		return NULL;

	gtk_tree_model_get (model, &iter, 0, &vcard, -1);
	g_assert (vcard != NULL && MIMEDIR_IS_VCARD (vcard));

	return vcard;
}


void
cardlist_widget_clear (CardListWidget *widget)
{
	GtkTreeModel *model;

	g_return_if_fail (widget != NULL);
	g_return_if_fail (CARDLIST_IS_WIDGET (widget));

	model = gtk_tree_view_get_model (GTK_TREE_VIEW (widget));

	gtk_list_store_clear (GTK_LIST_STORE (model));
}


void
cardlist_widget_set_list (CardListWidget *widget, GList *list)
{
	g_return_if_fail (widget != NULL);
	g_return_if_fail (CARDLIST_IS_WIDGET (widget));

	cardlist_widget_clear (widget);
	cardlist_widget_append_list (widget, list);
}


void
cardlist_widget_append_list (CardListWidget *widget, GList *list)
{
	GtkTreeModel *model;

	g_return_if_fail (widget != NULL);
	g_return_if_fail (CARDLIST_IS_WIDGET (widget));

	model = gtk_tree_view_get_model (GTK_TREE_VIEW (widget));

	for (; list != NULL; list = g_list_next (list)) {
		if (MIMEDIR_IS_VCARD (list->data))
			cardlist_widget_internal_append (widget, MIMEDIR_VCARD (list->data));
		else
			g_critical ("List item is not of type MIMEDirVCard");
	}
}


void
cardlist_widget_append_card (CardListWidget *widget, MIMEDirVCard *card)
{
	GtkTreeModel     *model;
	GtkTreePath      *path;
	GtkTreeIter       iter;
	GtkTreeSelection *selection;

	g_return_if_fail (widget != NULL);
	g_return_if_fail (CARDLIST_IS_WIDGET (widget));
	g_return_if_fail (card != NULL);
	g_return_if_fail (MIMEDIR_IS_VCARD (card));

	model     = gtk_tree_view_get_model (GTK_TREE_VIEW (widget));
	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (widget));

	iter = cardlist_widget_internal_append (widget, card);

	path = gtk_tree_model_get_path (model, &iter);
	gtk_tree_view_scroll_to_cell (GTK_TREE_VIEW (widget),
				      path, NULL,
				      TRUE, 0.5, 0.0);

	gtk_tree_selection_select_path (selection, path);
	gtk_tree_path_free (path);
}


void
cardlist_widget_remove_card (CardListWidget *widget, MIMEDirVCard *card)
{
	g_return_if_fail (widget != NULL);
	g_return_if_fail (CARDLIST_IS_WIDGET (widget));
	g_return_if_fail (card != NULL);
	g_return_if_fail (MIMEDIR_IS_VCARD (card));

	cardlist_widget_internal_remove (widget, card);
}


gboolean
cardlist_widget_open (CardListWidget *widget, const gchar *filename, GError **error)
{
	g_return_val_if_fail (widget != NULL, FALSE);
	g_return_val_if_fail (CARDLIST_IS_WIDGET (widget), FALSE);
	g_return_val_if_fail (filename != NULL, FALSE);

	cardlist_widget_clear (widget);

	return cardlist_widget_append (widget, filename, error);
}


gboolean
cardlist_widget_append (CardListWidget *widget, const gchar *filename, GError **error)
{
	GError *err = NULL;
	GtkTreeModel *model;
	GList *list, *current;
        
	g_return_val_if_fail (widget != NULL, FALSE);
	g_return_val_if_fail (CARDLIST_IS_WIDGET (widget), FALSE);
	g_return_val_if_fail (filename != NULL, FALSE);

	list = mimedir_vcard_read_list (filename, &err);
	if (err) {
		g_propagate_error (error, err);
		return FALSE;
	}

	model = gtk_tree_view_get_model (GTK_TREE_VIEW (widget));

	for (current = list; current != NULL; current = g_list_next (current)) {
		g_assert (current->data != NULL && MIMEDIR_IS_VCARD (current->data));
		cardlist_widget_internal_append (widget, MIMEDIR_VCARD (current->data));
	}

	mimedir_vcard_free_list (list);

	return TRUE;
}


gboolean
cardlist_widget_save (CardListWidget *widget, const gchar *filename, GError **error)
{
	GError *err = NULL;
	GList *list = NULL;
	GtkTreeIter iter;
	GtkTreeModel *model;

	g_return_val_if_fail (widget != NULL, FALSE);
	g_return_val_if_fail (CARDLIST_IS_WIDGET (widget), FALSE);
	g_return_val_if_fail (filename != NULL, FALSE);

	model = gtk_tree_view_get_model (GTK_TREE_VIEW (widget));

	if (gtk_tree_model_get_iter_root (model, &iter)) {
		do {
			MIMEDirVCard *vcard;

			gtk_tree_model_get (model, &iter, 0, &vcard, -1);
			g_assert (vcard != NULL && MIMEDIR_IS_VCARD (vcard));

			list = g_list_append (list, vcard);
		} while (gtk_tree_model_iter_next (model, &iter));
	}

	if (!mimedir_vcard_write_list (filename, list, &err)) {
		g_propagate_error (error, err);
		return FALSE;
	}

	return TRUE;
}
